#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

The RAMSES-II protocol (messages, packets/commands).
"""

__version__ = "0.20.19"
VERSION = __version__
